var canvas = null;
var canvasRect;
var lastPos = null;
var line = null;
var stroke = null;
var lineset = new Array();
var currentPos = 0;
var buttonListeners = new Array();

function tr( key ) {
    try {
        return localizedStrings[ key ];
    } catch ( e ) {
        debug( e );
        return key;
    }
}

function prefInt( key, defval ) {
    if ( window.widget ) {
        try {
            var val = widget.preferenceForKey( key );
//            return parseInt( val );
            if ( val ) {
                //if ( val && val != null && val.matches( /^\d+$/ ) ) {//val != undefined && val != NaN && val != null ) {
                return parseInt( val );
            }
        } catch ( e ) {
            debug( e );
        }
    }
    return defval;
}

var GifStatus = new function() {
    var tm = "" + (new Date()).getTime();
    this.id        = tm.substring( tm.length - 6, tm.length );
    this.delay     = prefInt( 'delay', 6 );
    this.frames    = prefInt( 'frames', 8 );
    this.vibration = prefInt( 'wobble', 4 );
    this.skip      = prefInt( 'sampling', 4 );
    this.lineWidth = prefInt( 'linewidth', 2 );
    this.shadow    = { position:[ 2,5 ], color:[ 200, 200, 200 ], 
                       display:true };
    this.filename  = "/tmp/animgif" + this.id + ".gif";
    this.animating = false;

    if ( window.widget ) {
        this.shadow.display = ( prefInt( 'shadow', 1 ) == 1 );
    }

    this.toString = function() {
        try {
            return widget.preferenceForKey( 'wobble' ) + " " + widget.preferenceForKey( 'sampling' );
        } catch ( e ) {
            return "Delay " + this.delay + ", Wobble " + this.vibration + ", Line " + this.lineWidth + ", Sampling " + this.sampling + ", Shadow " + this.shadow.display + ", Frames " + this.frames + e;
        }
    }
}

function debug( message ) {
    document.getElementById( "debugDiv" ).innerText = message;
}

function Line( color, width ) {
    this.color = "block";
    this.width = width;
    this.data  = new Array();
    this.push  = _Line__push;
    this.draw  = _Line__draw;
    this.write = _Line__write;
    this.toString = _Line__toString;
}

function _Line__push( x, y ) {
    var point = [ x, y ];
    this.data.push( point );
}

function _Line__toString() {
    return "Line : " + this.width + " " + this.color + " " + this.data.length;
}

function _Line__draw( context ) {
    var data = this.data;
    context.lineWidth = this.width;
    context.fillStyle = "black";
    context.moveTo( data[ 0 ][ 0 ], data[ 0 ][ 1 ] );
    for ( var i = 1; i < data.length; i++ ) {
        context.lineTo( data[ i ][ 0 ], data[ i ][ 1 ] );
    }
    context.stroke();
    context.save();
}

function _Line__write( command, skip ) {
    if ( skip == undefined ) skip = 1;
    command.write( "c,0,0,0\n" );
    command.write( "w," + this.width  + "\n");
    command.write( "l" );
    for ( var i = 0; i < this.data.length - 1; i+= skip ) {
        command.write( "," + this.data[ i ][ 0 ] + "," + this.data[ i ][ 1 ] );
    }
    command.write( "," + this.data[ this.data.length - 1 ][ 0 ] 
                   + "," + this.data[ this.data.length - 1 ][ 1 ] );
    command.write( "\n" );
}

function setup() {
    setCanvas( document.getElementById( "canvas" ) );
    if ( window.widget ) {
        widget.onhide   = hideWidget;
        widget.onremove = removeTemporaryFile; 
    }

    
    // buttons
    buttonListeners[ 'clear' ] 
        = setButtonAction( document.getElementById( 'clearButton' ), 
                           clearCanvas ); 
    buttonListeners[ 'stop' ] 
        = setButtonAction( document.getElementById( 'stopButton' ), 
                           stopMotion );
    buttonListeners[ 'animation' ] 
        = setButtonAction( document.getElementById( 'animButton' ), 
                           makeMotion );
    buttonListeners[ 'redo' ] 
        = setButtonAction( document.getElementById( 'redoButton' ), 
                         redoLine );
    buttonListeners[ 'undo' ] 
        = setButtonAction( document.getElementById( 'undoButton' ), undoLine )
;
    // menus
    setLineWidthMenu();
    setWobbleMenu();
    setShadowMenu();
    setSamplingMenu();
    setDelayMenu();

    createGenericButton( document.getElementById( 'done' ),
                         tr( "Done" ), flipToFront );

    setButtonAvailability();

    debug( GifStatus );
}

function setButtonAvailability() {
    try {
        buttonListeners[ 'clear' ].enable( ( currentPos > 0 ) );
        buttonListeners[ 'stop' ].enable( GifStatus.animating );
        buttonListeners[ 'animation' ].enable( !(GifStatus.animating ) && currentPos > 0 );
        buttonListeners[ 'redo' ].enable( ( currentPos < lineset.length ) );
        buttonListeners[ 'undo' ].enable( ( currentPos > 0 ) );

    } catch ( e ) {
        debug( e );
    }
}

function setCanvas( canvasElement ) {
    canvas = canvasElement;
    var cstyle = document.defaultView.getComputedStyle( canvas );
    canvasRect = { left:parseInt( cstyle.getPropertyValue( 'left' ) ),
                   top:parseInt( cstyle.getPropertyValue( 'top' ) ),
                   width:parseInt( cstyle.getPropertyValue( 'width' ) ),
                   height:parseInt( cstyle.getPropertyValue( 'height' ) ) };
    var context = canvas.getContext( "2d" );
    debug( canvasRect.width + ", " + canvasRect.height + " : " + context.fillStyle + ' , ' + context.lineWidth );
    context.fillStyle = "black";
    context.lineWidth = parseFloat( GifStatus.lineWidth );
    context.clearRect( 0, 0, canvasRect.width, canvasRect.height );
    context.save();
}

function canvasMouseDown( event, targetCanvas ) {
    if ( targetCanvas == undefined ) {
        if ( canvas ) {
            targetCanvas = canvas;
        } else {
            setCanvas( document.getElementById( "canvas" ) );
            if ( targetCanvas == null ) {
                throw new Exception( "no canvas" );
            }
        }
    }
    if ( canvas != targetCanvas ) {
        setCanvas( targetCanvas );
    }

    var x = event.x - canvasRect.left;
    var y = event.y - canvasRect.top;
    lastPos = { x:x, y:y };

    var context = canvas.getContext( "2d" );
    context.lineWidth = GifStatus.lineWidth;
    stroke = new Line( context.fillStyle, parseFloat( GifStatus.lineWidth ) );

    window.addEventListener( 'mousemove', canvasMouseMove, true );
    window.addEventListener( 'mouseup', canvasMouseUp, true );
    event.stopPropagation();
    event.preventDefault();
}

function canvasMouseMove( event ) {
    var context = canvas.getContext( "2d" );
    var x = event.x - canvasRect.left;
    var y = event.y - canvasRect.top;
    context.moveTo( lastPos.x, lastPos.y );
    context.lineTo( x, y );
    context.stroke();
    context.save();
    lastPos = {x:x, y:y };
    stroke.push( x, y );
    debug( stroke + " " + GifStatus.lineWidth);
}

function canvasMouseUp( event ) {
    if ( lineset.length > currentPos ) {
        lineset = lineset.slice( 0, currentPos );
    }

    lineset.push( stroke );
    currentPos ++;
    window.removeEventListener( 'mousemove', canvasMouseMove, true );
    window.removeEventListener( 'mouseup', canvasMouseUp, true );
    event.stopPropagation();
    event.preventDefault();
    setButtonAvailability();
}



function setLineWidth( width ) {
   var context = canvas.getContext( "2d" );
   context.lineWidth = width;
   context.save();
}

function redraw( lines, start, end ) {
    if ( lines == undefined ) lines = lineset;
    if ( start == undefined ) start = 0;
    if ( end == undefined ) end = lines.length;

    var context = canvas.getContext( "2d" );
    context.clearRect( 0, 0, canvasRect.width, canvasRect.height );
    for ( var i = start; i < end; i++ ) {
        var line = lines[ i ];
        line.draw( context );
    }
    context.save();
    context.restore();
    setButtonAvailability();
}

function undoLine( event ) {
    debug( "undo" );
    if ( currentPos > 0 ) {
        currentPos --;
        redraw( lineset, 0, currentPos );
        debug( "undo finished" );
    } else {
        debug( "no more undo buffer" );
    }
    setButtonAvailability();
}

function redoLine( event ) {
    debug( "redo" );
    if ( GifStatus.animating ) {
        stopMotion();
    }
    if ( currentPos < lineset.length ) {
        currentPos ++;
        redraw( lineset, 0, currentPos );
        debug( "redo finished" );
    } else {
        debug( "no more buffer" );
    }
    setButtonAvailability();    
}

function endHandler( com ) {
    var output = com.outputString;
    if ( output && output.length > 0 ) {
        var anim = document.getElementById( 'animationPanel' );
        var uri = com.outputString.split( "\n" )[ 0 ] + "?" + ( new Date() ).getTime();
        anim.src = uri;
        anim.style.display = "block";
        GifStatus.animating = true;
    } else {
        debug( "bad output '" + output + "': \n " + com.errorString );
    }
    setButtonAvailability();
}

function stopMotion( event ) {
    var anim = document.getElementById( 'animationPanel' );
    anim.style.display = "none";
    GifStatus.animating = false;
    setButtonAvailability();
}

function clearCanvas( event ) {
    if ( GifStatus.animating ) {
        stopMotion();
    }
    if ( canvas ) {
        var context = canvas.getContext( "2d" );
        context.clearRect( 0, 0, canvasRect.width, canvasRect.height );
        context.save();
        lineset = new Array();
    }
    setButtonAvailability();
}

function hideWidget( event ) {
    stopMotion( event );
}

function removeTemporaryFile( event ) {
    if ( window.widget ) {
        widget.system( "/bin/rm -f " + GifStatus.filename );
    }
}

function makeMotion( event ) {
    debug( "motion" );
    if ( window.widget ) {
        var arg_size   = " -s " + canvasRect.width + "x" + canvasRect.height;
        var arg_out    = " -o " + GifStatus.filename;

        var arg_shadow = "";
        if ( GifStatus.shadow && GifStatus.shadow.display ) {
            arg_shadow = " -shadow "; 
            var scol = GifStatus.shadow.color;
            for ( var i = 0; i < 3; i++ ) {
                if ( i != 0 ) arg_shadow += ",";
                arg_shadow += scol[ i ];
            }
            arg_shadow += " -spos " + GifStatus.shadow.position[ 0 ]
                + "," + GifStatus.shadow.position[ 1 ];
        }
        

        var arg_vibration = " -v " + GifStatus.vibration + "x" + GifStatus.vibration;
        var arg_frames = " -f " + GifStatus.frames;
        var arg_delay  = " -d " + GifStatus.delay;
        var pycommand  = "/usr/bin/python drawgif.py ";

        pycommand += arg_shadow;
        pycommand += arg_size;
        pycommand += arg_vibration;
        pycommand += arg_out;
        pycommand += arg_frames;
        pycommand += arg_delay;

        debug( pycommand );

        var command = widget.system(  pycommand, endHandler );

        for ( var i = 0; i < lineset.length; i++ ) {
            lineset[ i ].write( command, GifStatus.skip );
        }
        command.close();
    }
}


function imageDragStart( event ) {
    event.dataTransfer.effectAllowed = "copy";
    event.dataTransfer.setData( "URL", document.getElementById( 'animationPanel' ).src );
}

function flipToBehind( event ) {
    if ( window.widget ) {
        var front  = document.getElementById( "front" ); 
        var behind = document.getElementById( "behind" ); 
        widget.prepareForTransition("ToBack"); 
        behind.style.display  = "block"; 
        front.style.display = "none"; 
        setTimeout ('widget.performTransition();', 0);  
    }
}

function flipToFront( event ) {
    var front  = document.getElementById( "front" ); 
    var behind = document.getElementById( "behind" ); 
    if ( window.widget ) widget.prepareForTransition("ToBack"); 
    behind.style.display  = "none"; 
    front.style.display = "block"; 
    
    var values = new Array();
    var items = [ 'linewidth', 'wobble', 'delay', 'sampling', 'shadow'];
    for  ( var i = 0; i < items.length; i++ ) {
        
        var value = parseInt( document.getElementById( items[ i ] ).value );
        values[ items[ i ] ] = value;
    }
    GifStatus.skip = values[ 'sampling' ];
    GifStatus.vibration = values[ 'wobble' ];
    GifStatus.delay = values[ 'delay' ];
    GifStatus.shadow.display = ( values[ 'shadow' ] == 1 );
    GifStatus.lineWidth = values[ 'linewidth' ];

    if ( window.widget ) {

        widget.setPreferenceForKey( '' + GifStatus.delay, 'delay' );
        widget.setPreferenceForKey( '' + parseInt( GifStatus.lineWidth ), 'linewidth' );
        widget.setPreferenceForKey( '' + GifStatus.frames, 'frames' );
        widget.setPreferenceForKey( '' + GifStatus.vibration, 'wobble' );
        widget.setPreferenceForKey( '' + GifStatus.skip, 'sampling' );
        if ( GifStatus.shadow.display ) {
            widget.setPreferenceForKey( '1' , 'shadow' );
        } else {
            widget.setPreferenceForKey( '0' , 'shadow' );
        }

        debug( "linewidth : " + '' + parseInt( GifStatus.lineWidth ) );

        // transit to front and repaint graphics
        setTimeout ('widget.performTransition();redraw();', 0);  
    }
}


/// set preference panel
function setSelect( item ) {
    var label = document.getElementById( item.labelId );
    if ( label ) {
        label.innerText = tr( item.label );
    }
    var select = document.getElementById( item.selectId );
    select.innerHTML = "";
    for ( var i = 0; i < item.values.length; i++ ) {
        var option = document.createElement( 'option' );
        option.setAttribute( 'value', item.values[ i ] );
        option.appendChild( document.createTextNode( tr( item.labels[ i ] ) ) );
        if ( item.defaultValue == item.values[ i ] ) {
            option.setAttribute( 'selected' );
        }
        select.appendChild( option );
    }
}

function setLineWidthMenu() {
    var item = { labelId: "LineWidthLabel", label:"Line Width", 
                 selectId: "linewidth", 
                 defaultValue:GifStatus.lineWidth,
                 values: [ 8, 4, 2, 1 ],
                 labels: [ "Thicker", "Thick", "Middle", "Thin" ] };
    setSelect( item );
}

function setWobbleMenu() {
    var item = { labelId: "WobbleLabel", label:"Wobbling", 
                 selectId: "wobble", 
                 defaultValue:GifStatus.vibration,
                 values: [ 10, 7, 4, 3 ],
                 labels: [ "Larger", "Large", "Middle", "Small" ] };//, "Smaller" ] };
    setSelect( item );
}

function setShadowMenu() {
    var defval;
    if ( GifStatus.shadow.display ) defval = 1;
    else defval = 0;
    var item = { labelId: "ShadowLabel", label:"Shadow", 
                 selectId: "shadow", 
                 defaultValue:defval,
                 values: [ 1, 0 ],
                 labels: [ "With Shadow", "No Shadow" ] }
    setSelect( item );
}

function setSamplingMenu() {
    var item = { labelId: "SamplingLabel", label:"Sampling", 
                 selectId: "sampling", 
                 defaultValue:GifStatus.skip,
                 values: [ 10, 7, 4, 2, 1 ],
                 labels: [ "Coaser", "Coase", "Middle", "Fine", "Finer" ] }
    setSelect( item );
}

function setDelayMenu() {
    var item = { labelId: "DelayLabel", label:"Delay", 
                 selectId: "delay", 
                 defaultValue:GifStatus.delay,
                 values: [ 100, 50, 20, 6, 3 ],
                 labels: [ "Slower", "Slow", "Middle", "Fast", "Faster" ] }
    setSelect( item );
}



// buttons

function ButtonListener( element, target ) {
    this.element = element;
    this.target = target;
    this.enabled   = true;

    this.mouseDown = _ButtonListener__mouseDown;
    this.mouseUp   = _ButtonListener__mouseUp;

    this.enable    = _ButtonListener__enable;

    var self = this;
    element.addEventListener( 'mousedown', function() { self.mouseDown(); } );
    element.addEventListener( 'mouseup', function() { self.mouseUp(); } );
    
}

function _ButtonListener__enable( flag ) {
    if ( flag == this.enabled ) return;
    this.enabled = flag;
    if ( flag ) {
        this.element.style.color = "#400";
    } else {
        this.element.style.color = "#c84";
    }
}

function _ButtonListener__mouseDown( event ) {
    if ( this.enabled == false ) return;
    this.element.style.background = "url(Images/Button2.png)";
}

function _ButtonListener__mouseUp( event ) {
    if ( this.enabled == false ) return;
    this.element.style.background = "url(Images/Button.png)";
    this.target( event );
}

function setButtonAction( element, targetFunc ) {
    var listener = new ButtonListener( element, targetFunc );
    return listener;
}
